/* GIFTOPCX version 2.0
	 Translates GIF to 256 colour PCX
*/

#include "stdio.h"
#include "stdlib.h"
#include "dos.h"
#include "alloc.h"
#include "string.h"
#include "conio.h"

#define NO_CODE -1
#define pixels2bytes(n) ((n+7)/8)

#define SCREENWIDE    320  /* mode 13 screen dimensions */
#define SCREENDEEP    200
#define STEP          32   /* size of a step when panning */

#define HOME          0x4700 /* cursor control codes */
#define CURSOR_UP     0x4800
#define CURSOR_LEFT   0x4b00
#define CURSOR_RIGHT  0x4d00
#define END           0x4f00
#define CURSOR_DOWN   0x5000

typedef struct {
		char sig[6];
		unsigned int screenwidth,screendepth;
		unsigned char flags,background,aspect;
		} GIFHEADER;

typedef struct {
		unsigned int left,top,width,depth;
		unsigned char flags;
		} IMAGEBLOCK;

setvgapalette(char *p,int n,int b);
strmfe(char *nw,char *old,char *ext);
GetKey(void);
unpackgif(FILE *fp);
unpackimage(FILE *fp,char *buffer,int bits,int width,int depth,int flags);
char *farPtr(char *p,long l);
error(char *s);
showpicture(char *p,int width,int depth,int bits,int background,char *palette);

char GIFname[80];

main(int argc,char *argv[])
{
	 FILE *fp;

	 if(argc<=1) error("I need a path to a GIF file");
	 if((fp=fopen(argv[1],"rb"))==NULL)
		error("Error opening the file");
	 unpackgif(fp);
	 fclose(fp);
	 return (0);
}

/* this function is called whn the GIF decoder encounters an extension */

doextension(FILE *fp)
{
 int n,i;

 puts("Skipping an extension");
 do {
	if((n=fgetc(fp)) !=EOF){for(i=0;i<n;++i) fgetc(fp);}
 } while (n> 0 && n != EOF);
return(0);}

/* unpack a GIF file */
unpackgif(FILE *fp)
{
		GIFHEADER gh;
		IMAGEBLOCK iblk;
		char *p,palette[768];
		int b,c,bits,background,width,depth,flags;


												/* make sure it's a GIF file */
												if(fread((char *)&gh,1,sizeof(GIFHEADER),fp) != sizeof(GIFHEADER))
												 error("Error reading GIF header");

		/* get the screen dimensions */
		bits=(gh.flags & 0x0007)+1;
		background=gh.background;
		/* get colour map if there is one */
		if (gh.flags & 0x80) {
			 c = 3*(1<<((gh.flags & 7)+1));
			 if(fread(palette,1,c,fp) != c) error("Error reading colour map");
												 }

		/* step through the blocks */
		while((c=fgetc(fp))==',' || c=='!' || c==0)
		{
																		/* if it's an image block ... */
																		if(c==',')
																		{
																		 puts("Wait...unpacking an image");

			/* get the start of the image block */
			if(fread(&iblk,1,sizeof(IMAGEBLOCK),fp) != sizeof(IMAGEBLOCK))
																		 error("Error reading image block");

			/* get the image dimensions */
			width=iblk.width;
			depth=iblk.depth;

			/* get the local color map if there is one */
																		if(iblk.flags & 0x80)
																			{ b=3*(1<<((iblk.flags & 0x0007)+1));
																				if(fread(palette,1,b,fp) != c) error("Error reading local colour map");
																				bits=(iblk.flags & 0x0007)+1;
																			}

			/* get the initial code size */
			if((c=fgetc(fp))==EOF) error("Bad initial code size");
			flags=iblk.flags;
			if((p=(char *)farmalloc((long)width*(long)depth)) == NULL)
			 puts("Can't allocate a buffer");
			/* unpack the image */
			unpackimage(fp,p,c,width,depth,flags);

			/* show the picture */
			showpicture(p,width,depth,bits,background,palette);
			farfree(p);
			/* all done */
			return(0);
			}
																		/* otherwise, it's an extension */
																		else if(c=='!') doextension(fp);
			}
																		error("This file doesn't contain an image");
return(0);
}
/* unpack an LZW compressed image */
unpackimage(FILE *fp,char *buffer,int bits,int width,int depth,int flags)
{
int bits2,codesize,codesize2,nextcode,thiscode,oldtoken,currentcode;
int oldcode,bitsleft,blocksize,line=0,byte=0,pass=0;
char *p,*q,b[255],*u,*linebuffer;

static char firstcodestack[4096];
static char lastcodestack[4096];
static int codestack[4096];

 static int wordmasktable[] = { 0x0000,0x0001,0x0003,0x0007,
																	0x000f,0x001f,0x003f,0x007f,
																	0x00ff,0x01ff,0x03ff,0x07ff,
																	0x0fff,0x1fff,0x3fff,0x7fff };
		static int inctable[] = { 8,8,4,2,0 }; /* interlace increments */
		static int startable[] = { 0,4,2,1,0 }; /* interlace starts */

		p=q=b;
		bitsleft=8;
		if(bits<2 || bits>8) error("Bad symbol size");
		bits2= 1<<bits;
		nextcode=bits2+2;
		codesize2=1<<(codesize=bits+1);
		oldcode=oldtoken=NO_CODE;
		if((linebuffer=(char *)malloc(width))==NULL)
		 error("Can't allocate a line buffer");
		/* loop until something breaks */
		for(;;)
		 {
			if(bitsleft==8)
			 {
				if(++p>=q && (((blocksize=fgetc(fp))<1) ||
				 (q=(p=b)+fread(b,1,blocksize,fp))<(b+blocksize)))
					 { free(linebuffer);error("Unexpected end of file");}
				bitsleft = 0;
		 }
		thiscode= *p;
		if((currentcode=(codesize+bitsleft))<=8)
		 {
			*p>>=codesize;
			bitsleft=currentcode;
		 }
		else
		 {
				 if(++p>=q && (((blocksize=fgetc(fp))<1) ||
				 (q=(p=b)+fread(b,1,blocksize,fp))<(b+blocksize)))
					 { free(linebuffer);error("Unexpected end of file");
		 }
		thiscode |=*p<<(8-bitsleft);
		if(currentcode<=16) *p >>= (bitsleft=currentcode-8);
		else
		 {
			if(++p>=q && (((blocksize=fgetc(fp))<1) ||
				 (q=(p=b)+fread(b,1,blocksize,fp))<(b+blocksize)))
					 { free(linebuffer);error("Unexpected end of file");
		 }
		thiscode |=*p<<(16-bitsleft);
		*p >>= (bitsleft=currentcode-16);
	 }
 }
 thiscode &= wordmasktable[codesize];
 currentcode=thiscode;

 if(thiscode == (bits2+1)) break;
 if(thiscode>nextcode)
 {free(linebuffer);error("Bad code");}

 if(thiscode==bits2)
	{nextcode=bits2+2;codesize2=1<<(codesize=(bits+1));
	oldtoken=oldcode=NO_CODE;continue;}

 u=firstcodestack;

if(thiscode==nextcode)
{
 if(oldcode==NO_CODE)
 {free(linebuffer);error("Bad first code");}
 *u++=oldtoken;
 thiscode=oldcode;
}

while(thiscode>=bits2) {
 *u++=lastcodestack[thiscode];
 thiscode=codestack[thiscode];
}

oldtoken=thiscode;
do {
 linebuffer[byte++]=thiscode;
 if(byte>=width)
 {
	memcpy(farPtr(buffer,(long)line*(long)width),linebuffer,width);
	byte=0;
	/* check for interlaced image */
	if(flags & 0x40)
	 {line+=inctable[pass];
		if(line>=depth) line=startable[++pass];} else ++line;
	}
 if(u<=firstcodestack) break;
 thiscode=*--u;
 } while (1);

 if(nextcode<4096 && oldcode != NO_CODE)
	{ codestack[nextcode]=oldcode;
		lastcodestack[nextcode]=oldtoken;
		if(++nextcode>=codesize2 && codesize <12 ) codesize2=1<<++codesize;
	}
	oldcode=currentcode;
}
free(linebuffer);
return(0);
}


showpicture(char *p,int width,int depth,int bits,int background,char *palette)
{
 union REGS r;
 int c,i,n,x=0,y=0;

 r.x.ax=0x0013;
 int86(0x10,&r,&r);

 setvgapalette(palette,1<<bits,background);
 if(width>SCREENWIDE) n=SCREENWIDE;
 else n=width;

 do {
	for(i=0;i<SCREENDEEP;++i) {
	 c=y+i;
	 if(c>=depth) break;
	 memcpy(MK_FP(0xa000,SCREENWIDE*i),farPtr(p,(long)width*(long)c)+x,n);
	 }
	 c=GetKey();
	 switch(c) {
		 case CURSOR_LEFT:
				if((x-STEP)>0) x-=STEP;
				else x=0;
				break;
		case CURSOR_RIGHT:
				if((x+STEP+SCREENWIDE)<width) x+=STEP;
				else if(width>SCREENWIDE)
				 x=width-SCREENWIDE;
				else x=0;
				break;
		case CURSOR_UP:
				if((y-STEP)>0) y-=STEP;
				else y=0;
				break;
		case CURSOR_DOWN:
				if((y+STEP+SCREENDEEP)<depth) y+=STEP;
				else if(depth>SCREENDEEP)
					y=depth-SCREENDEEP;
				else y=0;
				break;
		case HOME:
				x=y=0;
				break;
		case END:
				if(width>SCREENWIDE)
				 x=width-SCREENWIDE;
				else x=0;
				if(depth>SCREENDEEP)
				 y=depth-SCREENDEEP;
				else y=0;
				break;
		}
	 } while(c!=27);

	 r.x.ax=0x0003;
	 int86(0x10,&r,&r);

	}

 /* get one extended key code */
 GetKey()
 {
		int c;

		c=getch();
		if(!(c&0x00ff)) c=getch() << 8;
		return(c);
 }

/* set the VGA palette and background */
setvgapalette(char *p,int n,int b)
{
	union REGS r;
	int i;

	outp(0x3c6,0xff);
	for(i=0;i<n;++i) {
		 outp(0x3c8,i);
		 outp(0x3c9,(*p++)>>2);
		 outp(0x3c9,(*p++)>>2);
		 outp(0x3c9,(*p++)>>2);
	}
	r.x.ax=0x1001;
	r.h.bh=b;
	int86(0x10,&r,&r);
}

strmfe(char *nw,char *old,char *ext)
{
 while(*old!=0 && *old!='.') *nw++=*old++;
 *nw++='.';
 while(*ext) *nw++=*ext++;
 *nw=0;
}

char *farPtr(char *p,long l) /* Return a far pointer p + l */
{
 unsigned int seg,off;

 seg = FP_SEG(p);
 off = FP_OFF(p);
 seg += (off/16);
 off &= 0x000f;
 off += (unsigned int)(l & 0x000fL);
 seg += (l/16L);
 p=(char *)MK_FP(seg,off);
 return(p);
}

error(char *s)
{ puts(s);exit(1);return(0);}
















    
   
    


